@@ -8,13 +8,11 @@ import os
 import sys
 import _yappi
 import pickle
-import marshal
 import threading
 try:
     from thread import get_ident        # Python 2
 except ImportError:
     from threading import get_ident     # Python 3
-from collections import defaultdict
 
 class YappiError(Exception): pass
 
@@ -137,6 +135,7 @@ of b() when called by a, just like yappi.
 PSTAT only expects to have the above dict to be saved.
 """
 def convert2pstats(stats):
+    from collections import defaultdict
     """
     Converts the internal stat type of yappi(which is returned by a call to YFuncStats.get())
     as pstats object.
@@ -241,30 +240,31 @@ class YStat(dict):
     Class to hold a profile result line in a dict object, which all items can also be accessed as
     instance attributes where their attribute name is the given key. Mimicked NamedTuples.
     """
-    _KEYS = ()
+    _KEYS = {}
 
     def __init__(self, values):
         super(YStat, self).__init__()
 
-        for i, key in enumerate(self._KEYS):
+        for key,i in self._KEYS.items():
             setattr(self, key, values[i])
 
     def __setattr__(self, name, value):
-        if name in self._KEYS:
-            self[self._KEYS.index(name)] = value
+        self[self._KEYS[name]] = value
         super(YStat, self).__setattr__(name, value)
 
 class YFuncStat(YStat):
     """
     Class holding information for function stats.
     """
-    _KEYS = ('name', 'module', 'lineno', 'ncall', 'nactualcall', 'builtin', 'ttot', 'tsub', 'index',
-        'children', 'ctx_id', 'tavg', 'full_name')
+    _KEYS = {'name':0, 'module':1, 'lineno':2, 'ncall':3, 'nactualcall':4, 'builtin':5, 'ttot':6, 'tsub':7, 'index':8, 'children':9, 'ctx_id':10, 'tavg':11, 'full_name':12}
 
     def __eq__(self, other):
         if other is None:
             return False
         return self.full_name == other.full_name
+    
+    def __ne__(self, other):
+        return not self == other
 
     def __add__(self, other):
 
@@ -284,7 +284,7 @@ class YFuncStat(YStat):
         return self
 
     def __hash__(self):
-        return self.index
+        return hash(self.full_name)
 
     def is_recursive(self):
         # we have a known bug where call_leave not called for some thread functions(run() especially)
@@ -327,8 +327,7 @@ class YChildFuncStat(YFuncStat):
     """
     Class holding information for children function stats.
     """
-    _KEYS = ('index', 'ncall', 'nactualcall', 'ttot', 'tsub', 'tavg', 'builtin', 'full_name',
-        'module', 'lineno', 'name')
+    _KEYS = {'index':0, 'ncall':1, 'nactualcall':2, 'ttot':3, 'tsub':4, 'tavg':5, 'builtin':6, 'full_name':7, 'module':8, 'lineno':9, 'name':10}
 
     def __add__(self, other):
         if other is None:
@@ -344,12 +343,18 @@ class YThreadStat(YStat):
     """
     Class holding information for thread stats.
     """
-    _KEYS = ('name', 'id', 'tid', 'ttot','sched_count',)
+    _KEYS = {'name':0, 'id':1, 'tid':2, 'ttot':3,'sched_count':4,}
 
     def __eq__(self, other):
         if other is None:
             return False
         return self.id == other.id
+    
+    def __ne__(self, other):
+        return not self == other
+    
+    def __hash__(self, *args, **kwargs):
+        return hash(self.id)
 
     def _print(self, out, columns):
         for x in sorted(columns.keys()):
@@ -370,41 +375,61 @@ class YThreadStat(YStat):
                 out.write(StatString(self.sched_count).rtrim(size))
         out.write(LINESEP)
 
-class YStats(list):
+class YStats(object):
     """
     Main Stats class where we collect the information from _yappi and apply the user filters.
     """
     def __init__(self):
         self._clock_type = None
+        self._as_dict = {}
+        self._as_list = []
 
     def get(self):
         self._clock_type = _yappi.get_clock_type()
-        return self.sort(DEFAULT_SORT_TYPE, DEFAULT_SORT_ORDER)
+        self.sort(DEFAULT_SORT_TYPE, DEFAULT_SORT_ORDER)
+        return self
 
     def sort(self, sort_type, sort_order):
-        super(YStats, self).sort(key=lambda stat: stat[sort_type], 
+        self._as_list.sort(key=lambda stat: stat[sort_type], 
             reverse=(sort_order==SORT_ORDERS["desc"]))
         return self
 
     def clear(self):
-        del self[:]
+        del self._as_list[:]
+        self._as_dict.clear()
 
     def empty(self):
-        return (len(self) == 0)
+        return (len(self._as_list) == 0)
 
     def __getitem__(self, key):
         try:
-            return super(YStats, self).__getitem__(key)
+            return self._as_list[key]
         except IndexError:
             return None
+        
+    def count(self, item):
+        return self._as_list.count(item)
+    
+    def __iter__(self):
+        return iter(self._as_list)
+    
+    def __len__(self):
+        return len(self._as_list)
+    
+    def pop(self):
+        item = self._as_list.pop()
+        del self._as_dict[item]
+        return item
 
     def append(self, item):
         # increment/update the stat if we already have it
-        for cstat in self:
-            if cstat == item:
-                cstat += item
-                return
-        super(YStats, self).append(item)
+        
+        existing = self._as_dict.get(item)
+        if existing:
+            existing += item
+            return
+        self._as_list.append(item)
+        self._as_dict[item] = item
 
     def _print_header(self, out, columns):
         for x in sorted(columns.keys()):
@@ -424,27 +449,43 @@ class YStats(list):
             if self.count(x) > 1:
                 return False
         return True
-
-class YChildFuncStats(YStats):
     
+
+class YStatsIndexable(YStats):
+    
+    def __init__(self):
+        super(YStatsIndexable, self).__init__()
+        self._additional_indexing = {}
+    
+    def clear(self):
+        super(YStatsIndexable, self).clear()
+        self._additional_indexing.clear()
+         
+    def pop(self):
+        item = super(YStatsIndexable, self).pop()
+        self._additional_indexing.pop(item.index, None)
+        self._additional_indexing.pop(item.full_name, None)
+        return item
+     
+    def append(self, item):
+        super(YStatsIndexable, self).append(item)
+        # setdefault so that we don't replace them if they're already there.
+        self._additional_indexing.setdefault(item.index, item)
+        self._additional_indexing.setdefault(item.full_name, item)
+
     def __getitem__(self, key):
         if isinstance(key, int):
-            for item in self:
-                if item.index == key:
-                    return item
-            return None
+            # search by item.index
+            return self._additional_indexing.get(key, None)
         elif isinstance(key, str):
-            for item in self:
-                if item.full_name == key:
-                    return item
-            return None
+            # search by item.full_name
+            return self._additional_indexing.get(key, None)
         elif isinstance(key, YFuncStat) or isinstance(key, YChildFuncStat):
-            for item in self:
-                if item.index == key.index:
-                    return item
-            return None
+            return self._additional_indexing.get(key.index, None)
 
-        return super(YChildFuncStats, self).__getitem__(key)
+        return super(YStatsIndexable, self).__getitem__(key)
+
+class YChildFuncStats(YStatsIndexable):
 
     def sort(self, sort_type, sort_order="desc"):
         sort_type = _validate_sorttype(sort_type, SORT_TYPES_CHILDFUNCSTATS)
@@ -472,8 +513,9 @@ class YChildFuncStats(YStats):
         for stat in self:
             stat.strip_dirs()
         return self
+    
 
-class YFuncStats(YStats):
+class YFuncStats(YStatsIndexable):
 
     _idx_max = 0
     _sort_type = None
@@ -484,21 +526,7 @@ class YFuncStats(YStats):
     def __init__(self, files=[]):
         super(YFuncStats, self).__init__()
         self.add(files)
-
-    def __getitem__(self, key):
-        if isinstance(key, int):
-            for item in self:
-                if item.index == key:
-                    return item
-            return None
-        elif isinstance(key, str):
-            for item in self:
-                if item.full_name == key:
-                    return item
-            return None
-
-        return super(YFuncStats, self).__getitem__(key)
-
+        
     def strip_dirs(self):
         for stat in self:
             stat.strip_dirs()
@@ -912,12 +940,19 @@ set_context_name_callback(None)
 
 def main():
     from optparse import OptionParser
-    usage = "yappi.py [-b] [-s] [scriptfile] args ..."
+    usage = "yappi.py [-b] [-o output_file] [-f output_format] [-s] [scriptfile] args ..."
     parser = OptionParser(usage=usage)
     parser.allow_interspersed_args = False
     parser.add_option("-b", "--builtins",
                   action="store_true", dest="profile_builtins", default=False,
                   help="Profiles builtin functions when set. [default: False]")
+    parser.add_option("-o", "--output-file", metavar="output_file",
+                  help="Write stats to output_file.")
+    parser.add_option("-f", "--output-format", default="pstat",
+                  choices=("pstat", "callgrind", "ystat"),
+                  metavar="output_format", help="Write stats in the specified"
+                  "format (\"pstat\", \"callgrind\" or \"ystat\", default is "
+                  "\"pstat\").")
     parser.add_option("-s", "--single_thread",
                   action="store_true", dest="profile_single_thread", default=False,
                   help="Profiles only the thread that calls start(). [default: False]")
@@ -931,15 +966,21 @@ def main():
     if (len(sys.argv) > 0):
         sys.path.insert(0, os.path.dirname(sys.argv[0]))
         start(options.profile_builtins, not options.profile_single_thread)
-        if sys.version_info >= (3, 0):
-            exec(compile(open(sys.argv[0]).read(), sys.argv[0], 'exec'),
-               sys._getframe(1).f_globals, sys._getframe(1).f_locals)
-        else:
-            execfile(sys.argv[0], sys._getframe(1).f_globals, sys._getframe(1).f_locals)
-        stop()
-        # we will currently use default params for these
-        get_func_stats().print_all()
-        get_thread_stats().print_all()
+        try:
+            if sys.version_info >= (3, 0):
+                exec(compile(open(sys.argv[0]).read(), sys.argv[0], 'exec'),
+                   sys._getframe(1).f_globals, sys._getframe(1).f_locals)
+            else:
+                execfile(sys.argv[0], sys._getframe(1).f_globals, sys._getframe(1).f_locals)
+        finally:
+            stop()
+            if options.output_file:
+                stats = get_func_stats()
+                stats.save(options.output_file, options.output_format)
+            else:
+                # we will currently use default params for these
+                get_func_stats().print_all()
+                get_thread_stats().print_all()
     else:
         parser.print_usage()
 
